#include <iostream>

#include <unordered_map>
#include <map>

#include <utility>  // std::forward

// Legend (for extremely simple generic functions)
// ===============================================
// K: key
// V: value
// C: container
// B: base (class)

// The 4th template argument should not be entered (must use the default), because it is an exploit
// used as a 'using B = C<K,V>' that can be defined before after template but before class definition
template <typename K, typename V, template <typename ... Args> class C = std::map, typename B = C<K,V> >
class MapWithDefault : private B
{
public:
    // Make default_value mandatory. Everything else follows the requested STL container
    template<typename... Args>
    MapWithDefault(V default_value, Args&& ... args) : B(std::forward<Args>(args)...), default_value(default_value) {};

public:
    using B::operator=;
    using B::get_allocator;

    using B::at;

    using B::operator[];

    // Read-only map (const object) uses only read-only operator[]
    const V& operator[](const K& key) const
    {
        auto it = this->find(key);
        return (it==this->end()) ? default_value : it->second;
    }

    using B::begin;
    using B::cbegin;
    using B::end;
    using B::cend;
    using B::rbegin;
    using B::crbegin;
    using B::rend;
    using B::crend;

    using B::empty;
    using B::size;
    using B::max_size;

    using B::clear;
    using B::insert;
    // using B::insert_or_assign;   // C++17
    using B::emplace;
    using B::emplace_hint;
    using B::erase;
    using B::swap;

    using B::count;
    using B::find;
    using B::equal_range;
    using B::lower_bound;
    using B::upper_bound;

public:
    const               V default_value;
    const MapWithDefault& read_only = static_cast<MapWithDefault&>(*this);
};


using namespace std;    // Demo (caller) code here, using a lot of cout

void print_rabbit(const auto& m)
{
    cout << "print_rabbit(with const arg ): " << m["rabbit"] << endl;
}

// void print_rabbit_unprotected(MapWithDefault<string,int>& m)
void print_rabbit_unprotected(auto& m)
{
    cout << "print_rabbit(unprotected arg): " << m["rabbit"] << endl;
}

int main()
{
    MapWithDefault<string, int> m(17);  // Endowed with default of 17
    // MapWithDefault<string, int, std::unordered_map> m(17);  // Endowed with default of 17
    auto& m_read_only = m.read_only;    // Usage: give read-only clients the read-only reference/alias to prevent writes

    // auto& m_read_only = m.read_only.read_only.read_only;    // Just to show that you can do weird stuff like this with no harm

    // Works as intended provided the users write const-correct arguments
    print_rabbit(m);                       // Should read 17 because the input arg applies const to it
    print_rabbit(m_read_only);
    print_rabbit_unprotected(m_read_only); // Should read 17 because the caller const protected it.
    print_rabbit_unprotected(m.read_only); // Should read 17 because the caller const protected it.

    // Directly operations on the read-only version (reference)
    cout << "pull rabbit from m.read_only:  " << m.read_only["rabbit"] << endl;   // Should read 17

    // Demonstrates commonly unwanted behavior of inserting requested key when not found
    cout << "pull rabbit from m:            " << m["rabbit"] << endl; // Should read 0 because the key was inserted (not default anymore)

    // Won't compile: demonstrate that it's read only
    // m_read_only["rabbit"] = 42;
    // m.read_only["rabbit"] = 42;

    // Demonstrate writing
    m["rabbit"] = 42;

    // Confirms written value
    cout << "pull rabbit from m_read_only:  " << m_read_only["rabbit"] << endl;   // Should read 42
    cout << "pull rabbit from m:            " << m["rabbit"] << endl;             // Should read 42

    return 0;
}
